'''Data loader module.

Data must be in the data directory of your game.

The following gamedata types are expected in the following subfolders
    - Music: data/bgm/
    - Sounds: data/sfx/
    - Graphics: data/gfx/

pygame.mixer and pygame.display must be initialized to load sound and graphics.

'''


import os
import pickle

import pygame

#path to data dir
__data_py__ = os.path.abspath(os.path.dirname(__file__))
__data_dir__ = os.path.normpath(os.path.join(__data_py__,'..','..','data'))

####
# taken from pyweek skellington
def filepath(*args):
    '''Determine the path to a file in the data directory.

    You can pass several arguments to this function for files in subfolders.

    For example: filepath("sfx", "kitten", "meow", "angry_meow.wav")

    Returns: "<gamedir>/data/sfx/kitten/meow/angry_meow.wav"
    '''
    return os.path.join(__data_dir__, *args)

def load(filename, mode='rb'):
    '''Open a file in the data directory.

    @param mode: sets the file mode can be any combination of
        - "r", meaning "read"
        - "w", meaning "write"
        - "b", meaning "binary" (useful for non-text files)
    '''
    return open(os.path.join(__data_dir__, filename), mode)

####

def get_filenames(path="", prefix="", suffix=""):
    """get all files in a folder

    @param path: path to folder
    @param prefix: prefix of requested files
    @param suffix: suffix of requested files (e.g. ".png")
    """

    #slicing magick and list filtering
    filenames = [elem for elem in os.listdir(path) \
        if elem.startswith(prefix) and elem.endswith(suffix)]
    return filenames


def load_image(filename, colorkey=None, alpha=False):
    """Loads image from filename

    @param colorkey: optional colorkey for transparency
    @param alpha: set alpha=True if the image already has transparency
    """
    path = os.path.join(__data_dir__, "gfx", filename)
    img = pygame.image.load(path)
    if colorkey:
        img.set_colorkey(colorkey)

    if alpha:
        return img.convert_alpha()
    else:
        return img.convert()

class NoneSound:
    """Dummy sound

    pretends to be a Sound object but doesn't do anything
    """
    def play(self, loops=0, maxtime=0, fade_ms=0):
        pass

    def stop(self):
        pass

    def fadeout(self, time):
        pass

    def set_volume(self, value):
        pass

    def get_volume(self):
        return 0

    def get_num_channels(self):
        return 0

    def get_length(self):
        return 0

    def get_buffer(self):
        return None

def load_sound(filename):
    """Loads and returns sound object

    Returns DummySound if something is wrong with audio or file
    not found.
    """
    path = os.path.join(__data_dir__, "sfx", filename)
    try:
        return pygame.mixer.Sound(path)
    except:
        print "Could not load sound: ", path
    return NoneSound()

def load_gfx(filenames, colorkey=(0,255,0,255), alpha=False):
    """

    @param filenames: a list of filenames
    @param colorkey: colorkey for tansparency
    @param alpha: set alpha=True if images have alpha channel
    @return: Dictionary with the loaded graphics. The filenames are used
    as the dictionary's keys
    """

    dc_gfx = {}
    for name in filenames:
        dc_gfx[name] = load_image(name, colorkey, alpha)
        dc_gfx[name].set_colorkey(colorkey)
    return dc_gfx

def load_sfx(filenames):
    """Load sound file

    @param filenames: list of filenames
    @return: Dictionary of sounds. The filenames are used as the
    dictionary's keys
    """
    sounds = {}
    for name in filenames:
        sounds[name] = load_sound(name)
    return sounds




def load_object(filename):
    """unpickles and returns object from file
    """
    f = None
    try:
        f = open(filepath(filename),"rb")
        object = pickle.load(f)
        f.close()
        return object
    except:
        if f:
            f.close()
        print "Could not unpickle ", filename
        return None


def save_object(object, filename):
    """pickles object to file in 'data/' directory
    """
    f = None
    try:
        f = open(filepath(filename),"wb")
        pickle.dump(object, f, 2)
        f.close()
        return True
    except:
        if f:
            f.close()
        print "Could not pickle ", object, " to file ", filename
        return False


def load_grid(filename, width=32, height=32, colorkey=None, alpha=False):
    """Cut image into several surfaces of (width, height) size

    Useful if you put a couple of graphics next to each other into one file
    (i.e. animations, tileset images, etc.).
    Note that these graphics must have the same size.

    Example:
        - tileset = load_grid("snowland.png")
        - tileset[1][2] would be the 2nd image in the 3rd row from that file


    @param filename: filename of image
    @return: a nested list with the surfaces cut from the image
    """

    img = load_image(filename, alpha)
    sw, sh = img.get_size()

    if alpha:
        img.covert_alpha()
    else:
        img.convert()


    images = []
    surf = pygame.Surface((width, height))
    y = 0
    while y < sh:
        x = 0
        line = []
        while x < sw:
            if colorkey:
                surf = pygame.Surface((width, height))
                surf.set_colorkey(colorkey)
            #blit big image on smaller surface with x and y as offset
            surf.blit(img, (-x, -y))

            line.append(surf.copy())
            x += width
        images.append(line)
        y += height
    return images



class Music:
    """Simple music class

    Saves the music path and plays it
    """
    def __init__(self, filename):
        try:
            self.path = os.path.join(__data_dir__, "bgm", filename)
        except:
            print "Could not load music: ", path
            self.path = None

    def play(self, rep=0):
        """play music
        """
        if self.path:
            pygame.mixer.music.stop()
            pygame.mixer.music.load(self.path)
            pygame.mixer.music.play(rep)

class JukeBox:
    """Handles music files, playback and stopping

    This is more like a basic music container than a jukebox.
    """
    def __init__(self, filenames):
        """JukeBox constructor

        music paths are saved in a dictionary with the filenames as keys

        @param filenames: list of music filenames
        """
        self.bgm = {}
        for name in filenames:
            self.bgm[name] = Music(name)

    def play(self, filename, rep=0):
        """play music, if filename is in Jukebox's dictionary
        """
        if filename in self.bgm.keys():
            self.bgm[filename].play(rep)

    def stop(self):
        """stop music playback
        """
        pygame.mixer.music.stop()
