'''Engine module

Contains state engine and basic state class
'''

import os
import sys
import string
import pygame
from pygame.locals import *

from event import *
from gui import Cursor
from data import get_filenames

WINDOWED = True
PRINTABLE = string.printable[:-5] #no \t \n  etc.

class Engine:
    """StateEgine

    Runs states, uses a list to stack states (when pausing a state)
    Attributes: states, current_state

    Example:
    e = Engine()  #initialise engine
    e.current_state = StartMenu()  #set starting state
    e.run()  #run your game

    That's basically all you need to know. Everything else will by done
    within the States.

    Confusing explanation of inner workings:
        1. Last state on state list is used as current_state.
        2. When current_state is done it is removed from the state list.
        3. If current_state is paused, it is kept on the state list.
        4. If current_state returns a state, this state is appended to
        state list to be run as NEXT current_state.
        5. If the state list is empty the game will be closed.

    Note that you can't jump between the stacked states.

    The last state in the state list will ALWAYS be run as current_state.
    """
    def __init__(self):
        self.current_state = State() #default state
        self.states = [self.current_state]  #state handling list

    def run(self):
        """Engine starts, runs current_state
        """
        self.states = [self.current_state]

        while self.states:
            self.current_state = self.states.pop()
            if self.current_state.paused:
                self.current_state.unpause()

            next, paused = self.current_state.mainloop()
            if self.current_state.kill_prev and self.states:
                self.states.pop()  #kill previous state too, if requested..

            if paused:  #paused states are kept
                self.states.append(self.current_state)

            if next: #set next state
                self.states.append(next)


class State:
    """State class.

    Important State attributes:
        - fps: frames per second (default: 20)
        - next: next state (must be initialized, i.e. self.next_state=GameOver())
        - kill_prev: if set to True, the previous (paused) state won't be resumed when this one quits
        - widgets: append all buttons (or other widgets) to that list to have them updated
        - events: add your keyboard or mouse events to this EventGroup (use add_kbevent for keyboard events)
        - screen: the window surface
        - background: a surface that is the same size as screen
    PRINT key creates screenshot
    F12 toggles fullscreen
    """
    def __init__(self):
        self.done = False
        self.next = None
        self.clock = pygame.time.Clock()
        self.events = EventGroup()
        self.widgets = []
        self.paused = False
        self.kill_prev = False
        self.fps = 20
        self.screen = pygame.display.get_surface()
        self.background = pygame.Surface(self.screen.get_size())
        self.sprites = pygame.sprite.Group()
        self._textentry = None
        self._addchar = None
        self._cursor = None

        self.add_kbevent(KEYUP, K_PRINT, self.make_screenshot)
        self.add_kbevent(KEYUP, K_F12, self.toggle_fullscreen)

    def pause(self):
        """should be called to pause state.

        Sets paused to True.
        """
        self.paused = True

    def unpause(self):
        """called when Engine returns to this state
        """
        self.paused = False
        self.done = False
        self.next = None

    def _main_start(self):
        """called just before the mainloop starts

        Sets mouse invisible if custom cursor set
        """
        if self._cursor:
            pygame.mouse.set_visible(0)
        else:
            pygame.mouse.set_visible(1)
        self.main_start()

    def main_start(self):
        """called just before the mainloop starts (empty placeholder)
        """
        pass


    def mainloop(self):
        """Mainloop

        Loop functions are called in this order:
            - check_events()
            - check_collisions()
            - update_screen()
            - update_other()
            - tick()

        @return: next, paused
        """
        self._main_start()
        while not self.done:
            self.check_events()
            self.check_collisions()
            self.update_screen()
            self.update_other()
            self.tick()
            pygame.event.pump()
        return self.next, self.paused

    def make_screenshot(self, prefix="screenshot-", suffix=".png"):
        """Saves an image of self.screen into working directory

        @param prefix: file prefix (default: "screenshot-")
        @param suffix: file type (default: ".png") ! do not forget the . !
        """

        screenshots = [int(elem[len(prefix):][:-len(suffix)])
            for elem in get_filenames(".", prefix, suffix)]

        if screenshots:
            nr = max(screenshots) + 1
        else:
            nr = 0

        filename = prefix + str(nr) + suffix
        pygame.image.save(self.screen, filename)
        #print filename, " saved."

    def check_events(self):
        """check for events, updates widgets
        """

        for ev in pygame.event.get():
            if ev.type == QUIT:
                self.close_game()
            for i in self.events:
                ev2 = i.ev
                #check keyboard events
                if ev.type==KEYUP or ev.type==KEYDOWN:
                    if ev.type == ev2.type and ev.key==ev2.key:
                        i.call()

                    #handle textentry
                    if self._textentry and ev.type==KEYUP and self._addchar:
                        self._textentry.add_text(char=self._addchar)
                        self._addchar = None

                    if self._textentry and ev.type==KEYDOWN and \
                        ev.unicode in PRINTABLE:
                        if self._addchar and self._addchar!=ev.unicode:
                            self._textentry.add_text(char=self._addchar)
                        self._addchar = ev.unicode


                #check mouse events
                elif ev.type==MOUSEBUTTONDOWN or ev.type==MOUSEBUTTONUP:
                    if ev.type==ev2.type and \
                        ev.button==ev2.button:
                        i.call()

        #update widgets
        for w in self.widgets:
            if w.update():
                w.draw(self.background)

    def add_kbevent(self, type, key, callback, **kwargs):
        """add Keyboard event

        @param type: KEYUP or KEYDOWN
        @param key: pygame key value (like K_r or K_SPACE)
        @param callback: function that should be called when key is pressed
        @param kwargs: keyword arguments passed to callback (optional)

        Example: self.add_kbevent(KEYDOWN, K_ESCAPE, self.quit)
        """
        ev = Event(type, {"key":key,"mod":KMOD_NONE}, callback=callback, **kwargs)
        self.events.add(ev)

    def add_textentry(self, textentry):
        """add TextEntry widget plus required keyboard events

        Currently you can have only ONE textentry in a state
        BACKSPACE deletes the last character
        RETURN calls TextEntry.submit()

        @param textentry: a TextEntry widget
        """

        #add widget
        self.widgets.append(textentry)
        self._textentry = textentry
        self.add_kbevent(KEYDOWN, K_BACKSPACE, textentry.rem_text)
        self.add_kbevent(KEYDOWN, K_RETURN, textentry.submit)
        return

    def set_cursor_image(self, image, clickpoint=(0,0)):
        """replace default mouse cursor with image

        Cursor variable is State._cursor, default cursor automatically
        hidden.

        @param image:a pygame.Surface
        @param clickpoint:  Usually that's the topleft or topright pixel of the image.
        If you're using a crosshair as cursor, the clickpoint would probably
        be the center.
        """
        self._cursor = Cursor(image, clickpoint)
        self.widgets.append(self._cursor)
        self.sprites.add(self._cursor)
        pygame.mouse.set_visible(0)

    def toggle_fullscreen(self):
        res = self.screen.get_size()
        global WINDOWED
        WINDOWED = not WINDOWED
        if WINDOWED:
            pygame.display.set_mode(res)
        else:
            pygame.display.set_mode(res, FULLSCREEN)
        self.screen.blit(self.background, (0,0))
        pygame.display.flip()

    def check_collisions(self):
        """check for sprite collisions (empty placeholder)
        """
        pass

    def update_screen(self):
        """update screen

        Use this method to handle screen updates any way you like.
        """
        self.screen.blit(self.background, (0,0))
        self.sprites.update()
        self.sprites.draw(self.screen)
        pygame.display.flip()

    def tick(self):
        """cap fps to self.fps
        """
        self.clock.tick(self.fps)

    def update_other(self):
        """update other stuff (empty placeholder)

        Use that method for other stuff that needs to be updated
        every frame.
        """
        pass

    def quit(self):
        """sets quit flag to True
        """
        self.done = True

    def close_game(self):
        """closes the entire game window
        """

        sys.exit(0)


