"""kitten (aka sprite) module

Sprite stuff that moves, is animated or any other kittenty
"""

import pygame

import utils

class Kitten(pygame.sprite.Sprite):
    """Basic sprite class

    Kittens preferred, although other images should work, too. ;)
    """
    def __init__(self, image):
        """Kitten constructor

        @param image: surface (usually a loaded image)
        """
        pygame.sprite.Sprite.__init__(self)
        self.image = image
        self.rect = self.image.get_rect()
        self.moving = False
        self.vel = (0,0)
        self.pos = (0,0)
        self.speed = 2

    def get_size(self):
        """return size of image
        """
        return self.image.get_size()

    def set_pos(self, pos):
        """set sprite position
        """
        self.pos = pos
        self.rect.topleft = pos

    def set_vel(self, vel):
        """set current velocity to vel

        The first value is the x-velocity (horizontal movement),
        the second is the y-velocity (vertical movement)
        So
            - (1,0) means "move right"
            - (-1,0) means "move left"
            - (1,2) means "move diagonally" ("right and a little more down")
        """
        self.vel = vel
        if self.vel[0] or self.vel[1]:
            self.moving = True
        else:
            self.moving = False

    def add_vel(self, vel):
        """add vel to current velocity
        """
        vx, vy = vel
        self.vel = (self.vel[0] + vx, self.vel[1] + vy)

    def update(self):
        """update sprite - add vel to sprite's position
        """
        if self.moving:
            self.set_pos((self.pos[0]+self.vel[0],self.pos[1]+self.vel[1]))

    def set_image(self, img):
        """change sprite's image
        """
        r = self.rect  #keep original position
        self.image = img
        self.rect = r
        self.set_pos(self.pos)


class AnimatedKitten(Kitten):
    """Animated Sprite class

    Again, kittens are preferred, although other animations work fine, too.
    """
    def __init__(self, images, delay=5):
        """AnimatedKitten constructor

        @param images: list of loaded images
        @param delay: delay between image changes
        """
        self.images = images
        self.cur_image = 0
        self.anim_timer = utils.Timer(delay)
        self.animate = False
        Kitten.__init__(self, images[self.cur_image])

    def set_images(self, images):
        """change images

        @param images: list of loaded images
        """
        self.images = images
        r = self.image.get_rect()
        self.cur_image = 0
        self.set_image(self.images[0])

    def update(self):
        """update AnimatedKitten

        move and/or animate, if needed
        """
        Kitten.update(self)
        if self.animate and self.anim_timer.update():
            self.cur_image += 1
            if self.cur_image >= len(self.images):
                self.cur_image = 0
            self.set_image(self.images[self.cur_image])

    def play(self):
        """show animation (always looped)
        """
        self.animate = True

    def stop(self):
        """animation stopps, first image displayed
        """
        self.animate = False
        self.cur_image = 0
        self.anim_timer.activate() #set timer to 0
        self.set_image(self.images[self.cur_image])

    def pause(self):
        """pause animation, last shown image is kept
        """
        self.animate = False

class Chara(AnimatedKitten):
    """animated 4-way moveable game character class

    Useful for classic top-down RPG characters
    """

    def __init__(self, imglists):
        """Chara constructor

        @param imglists: nested list of images, usually generated by data.load_grid
        """
        AnimatedKitten.__init__(self, imglists[0])
        self.imglists = imglists
        self.cur_list = 0
        self.movedir = None
        self.cur_dir = 0

    def set_movedir(self, dir):
        """set Chara's movedir

        @param dir:
            - 0 = down
            - 1 = up
            - 2 = left
            - 3 = right
        """
        self.movedir = dir
        self.play()

    def turn(self, dir):
        """turn chara to dir

        @param dir:
            - 0 = down
            - 1 = up
            - 2 = left
            - 3 = right
        """
        self.set_images(self.imglists[dir])
        self.cur_dir = dir

    def update(self):
        """update Chara

        move, animate, change imglists if Chara has turned..
        """
        if self.movedir is not None:
            dir = self.movedir
            if not self.moving or dir != self.cur_dir:
                self.cur_dir = dir
                self.set_images(self.imglists[dir]) #turn chara

                #set new vel
                if dir==0:  #down
                    self.set_vel((0, self.speed))
                elif dir==1:  #up
                    self.set_vel((0, -self.speed))
                elif dir==2:  #left
                    self.set_vel((-self.speed, 0))
                else:  #right
                    self.set_vel((self.speed, 0))
        elif not self.movedir and self.vel!=(0,0):
            self.stop_moving()
        AnimatedKitten.update(self)

    def stop_moving(self):
        """stop chara movement and animation
        """
        self.set_vel((0,0))
        self.movedir = None
        self.stop()

